#include "simulation.hpp"

using namespace cgp;

//#undef SOLUTION

#ifdef SOLUTION
void collision_sphere_plane(particle_structure& particle, cgp::vec3 const& n, cgp::vec3 const& p0)
{
	vec3 const& p = particle.p;
	vec3& v = particle.v;
	float const r = particle.r;
	float alpha = 0.95f;

	float const penetration = dot(p - p0, n) - r;
	if(penetration<0) {
		float velocity_component = dot(v, n);
		if(velocity_component<0) // need to bound
		{
			vec3 const vn = dot(v, n) * n;
			vec3 const vt = v - vn;
			v = alpha*(vt-vn);
		}
	}


}

void collision_sphere_sphere(particle_structure& particle_1, particle_structure& particle_2)
{
	vec3& p1 = particle_1.p;
	vec3& p2 = particle_2.p;
	vec3& v1 = particle_1.v;
	vec3& v2 = particle_2.v;
	float const r1 = particle_1.r;
	float const r2 = particle_2.r;

	float const alpha = 0.95f;

	vec3 const p12 = p1 - p2;
	float const d12 = norm(p12);

	if (d12 < r1 + r2)
	{
		vec3 const u12 = p12 / d12;
		
		float velocity_component = dot(v1-v2, p1-p2);
		if(velocity_component<0) {

			float const j = dot(v1 - v2, u12);
			v1 = v1 - alpha * j * u12;
			v2 = v2 + alpha * j * u12;
		}


	}
}

void cancel_colliding_velocity(particle_structure& particle_1, particle_structure& particle_2)
{
	vec3& p1 = particle_1.p;
	vec3& p2 = particle_2.p;
	vec3& v1 = particle_1.v;
	vec3& v2 = particle_2.v;

	float const r1 = particle_1.r;
	float const r2 = particle_2.r;

	vec3 const p12 = p1 - p2;
	float const d12 = norm(p12);

	if (d12 < r1 + r2)
	{
		vec3 const u12 = p12 / d12;

		float velocity_component = dot(v1-v2, u12);
		if(velocity_component<0) {
			v1 = v1-dot(v1,u12)*u12;
			v2 = v2-dot(v2,u12)*u12;
		}
	}

}
void cancel_sphere_plane_velocity(particle_structure& particle, cgp::vec3 const& n, cgp::vec3 const& p0)
{
	vec3 const& p = particle.p;
	vec3& v = particle.v;
	float const r = particle.r;

	float const penetration = dot(p - p0, n) - r;
	if(penetration<0) {
		float velocity_component = dot(v, n);
		if(velocity_component<0) // need to bound
		{
			v = v-dot(v,n)*n;
		}
	}
}
#endif

#ifdef SOLUTION
void simulate(std::vector<particle_structure>& particles, float dt_true)
{

	vec3 const g = { 0,0,-9.81f };
	size_t const N_substep = 10;
	float const dt = dt_true / N_substep;
	for (size_t k_substep = 0; k_substep < N_substep; ++k_substep)
	{
		size_t const N = particles.size();
		for (size_t k = 0; k < N; ++k)
		{
			particle_structure& particle = particles[k];

			vec3 const f = particle.m * g;

			particle.v = (1 - 0.9f * dt) * particle.v + dt * f / particle.m;
			
		}

		// Collisions between spheres
		for (size_t k1 = 0; k1 < N; ++k1)
			for (size_t k2 = k1 + 1; k2 < N; ++k2)
				collision_sphere_sphere(particles[k1], particles[k2]);
		

		// Collisions with cube
		numarray<vec3> face_normal = { {0, 1,0}, { 1,0,0}, {0,0, 1}, {0,-1,0}, {-1,0,0}, {0,0,-1} };
		numarray<vec3> face_position = { {0,-1,0}, {-1,0,0}, {0,0,-1}, {0, 1,0}, { 1,0,0}, {0,0, 1} };

		const size_t N_face = face_normal.size();
		for (size_t k = 0; k < N; ++k)
			for (size_t k_face = 0; k_face < N_face; ++k_face)
				collision_sphere_plane(particles[k], face_normal[k_face], face_position[k_face]);

		// forbid velocity going into other spheres
		for (size_t k1 = 0; k1 < N; ++k1)
			for (size_t k2 = k1 + 1; k2 < N; ++k2)
				cancel_colliding_velocity(particles[k1], particles[k2]);
		for (size_t k = 0; k < N; ++k)
			for (size_t k_face = 0; k_face < N_face; ++k_face)
				cancel_sphere_plane_velocity(particles[k], face_normal[k_face], face_position[k_face]);
				

		for (size_t k = 0; k < N; ++k)
		{
			particle_structure& particle = particles[k];
			particle.p = particle.p + dt * particle.v;
		}
	}

}
#else

void simulate(std::vector<particle_structure>& particles, float dt_arg)
{

	size_t const N_substep = 10;
	float const dt = dt_arg / N_substep;
	for (size_t k_substep = 0; k_substep < N_substep; ++k_substep)
	{

		vec3 const g = { 0,0,-9.81f };
		size_t const N = particles.size();

		// Update velocity with gravity force and friction
		for (size_t k = 0; k < N; ++k)
		{
			particle_structure& particle = particles[k];
			vec3 const f = particle.m * g;

			particle.v = (1 - 0.9f * dt) * particle.v + dt * f / particle.m;
		}

		// **************************************** //
		// TO DO: Collision Handling
		// 
		// The global pseudo code may be the following:
		//  (you need to implement steps 2 and 3)
		// 
		//   1: Velocity update with force fields (already implemented)
		//   - For all particles:
		//     Update velocity using gravity and friction 
		//  
		//   2: Impulse response for bouncing effect
		//   - For all particles:
		//     Check collision between particles 
		//      Then update their velocities using impulse response
		//   - For all particles:
		//     Check collision between cube faces 
		//      Then update their velocities using impulse response
		//
		//   3: Cancel velocity components in penetration
		//   - For all particles:
		//     Check collision between particles 
		//      Then cancel velocity components going inside other particle
		//   - For all particles:
		//     Check collision between cube faces 
		//      Then cancel velocity components going outside the cube
		//
		//   4: Position update (already implemented)
		// 	 - For all particles:
		//     Update position using the velocity
		//
		//  Hints:
		//    - The cube is by default centered around zero, faces are between [-1,1]
		//      To compute the intersection between sphere and cube faces, you may store
		//      the normals, and face center positions of the cube in a vectors.
		//
		//    - Implement and check first the collision between the spheres and the cube
		//      before the collision between spheres
		//
		// **************************************** //

		// Update position from velocity
		for (size_t k = 0; k < N; ++k)
		{
			particle_structure& particle = particles[k];
			particle.p = particle.p + dt * particle.v;
		}
	}

}
#endif